/**
 *****************************************************************************
 * @file       usb_device_custom_hid_if.c
 * @brief      Template of module implementation
 * @author     Temple8
 * @copyright  SDATAWAY
 *
 * History:
 * 2024.03.27  T8  Initial version
 *
 *****************************************************************************
 */
/* Includes -----------------------------------------------------------------*/
#include "usb_device_custom_hid_if.h"

#include <zephyr/init.h>

#include <zephyr/usb/usb_device.h>
#include <zephyr/usb/class/usb_hid.h>
#include <zephyr/sys/byteorder.h>
#include <zephyr/sys/util.h>
#include <usb_descriptor.h>

#include <zephyr/logging/log.h>

/* Global variables ---------------------------------------------------------*/

/* Private constants --------------------------------------------------------*/
#define LOG_LEVEL LOG_LEVEL_DBG
LOG_MODULE_REGISTER(usbCustomIf);

#define CUSTOM_81_EP_ADDR		0x81

#define CUSTOM_82_EP_ADDR		0x82

#define CUSTOM_83_EP_ADDR		0x83
#define CUSTOM_03_EP_ADDR		0x03

/* Private macros -----------------------------------------------------------*/

/* Private typedefs ---------------------------------------------------------*/
struct usb_hid_class_subdescriptor {
	uint8_t bDescriptorType;
	uint16_t wDescriptorLength;
} __packed;

struct usb_hid_descriptor {
	uint8_t bLength;
	uint8_t bDescriptorType;
	uint16_t bcdHID;
	uint8_t bCountryCode;
	uint8_t bNumDescriptors;

	/*
	 * Specification says at least one Class Descriptor needs to
	 * be present (Report Descriptor).
	 */
	struct usb_hid_class_subdescriptor subdesc[1];
} __packed;

struct usb_custom_if0_config {
	struct usb_if_descriptor if0;
	struct usb_hid_descriptor if0_hid;
	struct usb_ep_descriptor if0_81_ep;
} __packed;

struct usb_custom_if1_config {
	struct usb_if_descriptor if1;
	struct usb_hid_descriptor if1_hid;
	struct usb_ep_descriptor if1_82_ep;
} __packed;

struct usb_custom_if2_config {
	struct usb_if_descriptor if2;
	struct usb_hid_descriptor if2_hid;
	struct usb_ep_descriptor if2_83_ep;
	struct usb_ep_descriptor if2_03_ep;
} __packed;

USBD_CLASS_DESCR_DEFINE(primary, 0) struct usb_custom_if2_config custom_if2_cfg = {
	// Interface descriptor 2 
	.if2 = {
		.bLength = sizeof(struct usb_if_descriptor),
		.bDescriptorType = USB_DESC_INTERFACE,
		.bInterfaceNumber = 0,
		.bAlternateSetting = 0,
		.bNumEndpoints = 2,
		.bInterfaceClass = USB_BCC_HID,
		.bInterfaceSubClass = 0,
		.bInterfaceProtocol = 0,
		.iInterface = 0,
	},
  .if2_hid = {
		.bLength = sizeof(struct usb_hid_descriptor),
		.bDescriptorType = USB_DESC_HID,
		.bcdHID = sys_cpu_to_le16(USB_HID_VERSION),
		.bCountryCode = 0,
		.bNumDescriptors = 1,
		.subdesc[0] = {
			.bDescriptorType = USB_DESC_HID_REPORT,
			.wDescriptorLength = 30,
		},
	},	
	// Data Endpoint 83 
	.if2_83_ep = {
		.bLength = sizeof(struct usb_ep_descriptor),
		.bDescriptorType = USB_DESC_ENDPOINT,
		.bEndpointAddress = CUSTOM_83_EP_ADDR,
		.bmAttributes = USB_DC_EP_INTERRUPT,
		.wMaxPacketSize = sys_cpu_to_le16(64),
		.bInterval = 0x01,
	},
	// Data Endpoint 03 
	.if2_03_ep = {
		.bLength = sizeof(struct usb_ep_descriptor),
		.bDescriptorType = USB_DESC_ENDPOINT,
		.bEndpointAddress = CUSTOM_03_EP_ADDR,
		.bmAttributes = USB_DC_EP_INTERRUPT,
		.wMaxPacketSize = sys_cpu_to_le16(64),
		.bInterval = 0x01,
	},
};

USBD_CLASS_DESCR_DEFINE(primary, 0) struct usb_custom_if1_config custom_if1_cfg = {
	// Interface descriptor 1 
	.if1 = {
		.bLength = sizeof(struct usb_if_descriptor),
		.bDescriptorType = USB_DESC_INTERFACE,
		.bInterfaceNumber = 0,
		.bAlternateSetting = 0,
		.bNumEndpoints = 1,
		.bInterfaceClass = USB_BCC_HID,
		.bInterfaceSubClass = 0,
		.bInterfaceProtocol = 0,
		.iInterface = 0,
	},
  .if1_hid = {
		.bLength = sizeof(struct usb_hid_descriptor),
		.bDescriptorType = USB_DESC_HID,
		.bcdHID = sys_cpu_to_le16(USB_HID_VERSION),
		.bCountryCode = 0,
		.bNumDescriptors = 1,
		.subdesc[0] = {
			.bDescriptorType = USB_DESC_HID_REPORT,
			.wDescriptorLength = 84,
		},
	},	
	// Data Endpoint 82
	.if1_82_ep = {
		.bLength = sizeof(struct usb_ep_descriptor),
		.bDescriptorType = USB_DESC_ENDPOINT,
		.bEndpointAddress = CUSTOM_82_EP_ADDR,
		.bmAttributes = USB_DC_EP_INTERRUPT,
		.wMaxPacketSize = sys_cpu_to_le16(64),
		.bInterval = 0x01,
	},
};

USBD_CLASS_DESCR_DEFINE(primary, 0) struct usb_custom_if0_config custom_if0_cfg = {
	// Interface descriptor 0 
	.if0 = {
		.bLength = sizeof(struct usb_if_descriptor),
		.bDescriptorType = USB_DESC_INTERFACE,
		.bInterfaceNumber = 0,
		.bAlternateSetting = 0,
		.bNumEndpoints = 1,
		.bInterfaceClass = USB_BCC_HID,
		.bInterfaceSubClass = 0,
		.bInterfaceProtocol = 0,
		.iInterface = 0,
	},
  .if0_hid = {
		.bLength = sizeof(struct usb_hid_descriptor),
		.bDescriptorType = USB_DESC_HID,
		.bcdHID = sys_cpu_to_le16(USB_HID_VERSION),
		.bCountryCode = 0,
		.bNumDescriptors = 1,
		.subdesc[0] = {
			.bDescriptorType = USB_DESC_HID_REPORT,
			.wDescriptorLength = 101,
		},
	},	
	// Data Endpoint 81 
	.if0_81_ep = {
		.bLength = sizeof(struct usb_ep_descriptor),
		.bDescriptorType = USB_DESC_ENDPOINT,
		.bEndpointAddress = CUSTOM_81_EP_ADDR,
		.bmAttributes = USB_DC_EP_INTERRUPT,
		.wMaxPacketSize = sys_cpu_to_le16(64),
		.bInterval = 0x01,
	},
};

static struct usb_ep_cfg_data ep_83_cfg[] = {
	{
		.ep_cb = NULL,
		.ep_addr = CUSTOM_83_EP_ADDR,
	},	
	{
		.ep_cb = NULL,
		.ep_addr = CUSTOM_03_EP_ADDR,
	},
};

static struct usb_ep_cfg_data ep_82_cfg[] = {
	{
		.ep_cb = NULL,
		.ep_addr = CUSTOM_82_EP_ADDR,
	},
};

static struct usb_ep_cfg_data ep_81_cfg[] = {
	{
		.ep_cb = NULL,
		.ep_addr = CUSTOM_81_EP_ADDR,
	},
};

/* Private variables --------------------------------------------------------*/

/* Private prototypes -------------------------------------------------------*/

/* Public functions ---------------------------------------------------------*/

/* Private functions --------------------------------------------------------*/
static void custom_interface0_config(struct usb_desc_header *head,
				      uint8_t bInterfaceNumber)
{
	ARG_UNUSED(head);

	custom_if0_cfg.if0.bInterfaceNumber = bInterfaceNumber;
}

static void custom_interface1_config(struct usb_desc_header *head,
				      uint8_t bInterfaceNumber)
{
	ARG_UNUSED(head);

	custom_if1_cfg.if1.bInterfaceNumber = bInterfaceNumber;
}

static void custom_interface2_config(struct usb_desc_header *head,
				      uint8_t bInterfaceNumber)
{
	ARG_UNUSED(head);

	custom_if2_cfg.if2.bInterfaceNumber = bInterfaceNumber;
}

USBD_DEFINE_CFG_DATA(usb_custom_if2_config) = {
	.usb_device_description = NULL,
	.interface_config = custom_interface2_config,
	.interface_descriptor = &custom_if2_cfg.if2,
	.cb_usb_status = NULL,
	.interface = {
		.class_handler = NULL,
		.custom_handler = NULL,
		.vendor_handler = NULL,
	},
	.num_endpoints = ARRAY_SIZE(ep_83_cfg),
	.endpoint = ep_83_cfg,
};

USBD_DEFINE_CFG_DATA(usb_custom_if1_config) = {
	.usb_device_description = NULL,
	.interface_config = custom_interface1_config,
	.interface_descriptor = &custom_if1_cfg.if1,
	.cb_usb_status = NULL,
	.interface = {
		.class_handler = NULL,
		.custom_handler = NULL,
		.vendor_handler = NULL,
	},
	.num_endpoints = ARRAY_SIZE(ep_82_cfg),
	.endpoint = ep_82_cfg,
};

USBD_DEFINE_CFG_DATA(usb_custom_if0_config) = {
	.usb_device_description = NULL,
	.interface_config = custom_interface0_config,
	.interface_descriptor = &custom_if0_cfg.if0,
	.cb_usb_status = NULL,
	.interface = {
		.class_handler = NULL,
		.custom_handler = NULL,
		.vendor_handler = NULL,
	},
	.num_endpoints = ARRAY_SIZE(ep_81_cfg),
	.endpoint = ep_81_cfg,
};